

package GravityWars;

import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;

class Astronave
{
	int maxX = 640 , maxY = 480;
  double centroX,centroY;

	// Direzione: angolo rispetto a 0Pi
	double angolo;
  double rotationDelta;

	// Posizione del baricentro dell'astronave
	int x , y;

	// Componenti della velocita' dell'astronave
	double vel_x;
	double vel_y;

	// Massima velocita'
	double MAX_VEL = 10D;

  // Massima accelerazione
	double MAX_ACC = 999999D;

	// Accelerazione dell'astronave
	double deltaVel;

	// Accelerazione di gravita'
	static double grav = 5000D;
  double shipMass = 1D;
	static double attr = 0.01D;

	// Fill o meno dell'astronave
	boolean fill;

  // Posizione standard dell'astronave
  int standardX,standardY;

	// Forma dell'astronave e relativo sprite
	Polygon shape;
	Polygon sprite;

	public Astronave( Polygon p ,int x ,int y , boolean fill ) {
		shape = p;
		standardX = this.x = x;
		standardY = this.y = y;
		vel_x=0;
		vel_y=0;
		deltaVel=0;
		angolo = 0;
    this.fill = fill;
    sprite = new Polygon();

	}

	public Astronave( int posX , int posY , boolean fill) {

		// Astronave standard (un triangolo)
		shape = new Polygon();
		shape.addPoint( -9 , -8 );
    shape.addPoint( -9 , 0 );
		shape.addPoint( -9 , 8 );
		shape.addPoint(  10 , 0 );

    // Posizioni standard
		standardX = x = posX; standardY = y = posY;
    centroX = (double)(maxX/2D);
    centroY = (double)(maxY/2D);

		restart();
		this.fill = fill;
    sprite = new Polygon();

	}

  public void setArea( int x , int y ) {
    maxX = x;
    maxY = y;
  }

  public void setGravityCenter( double x , double y ) {
    centroX = x;
    centroY = y;
  }

  public static void setCenterMass( double mass ) {
    grav = mass;
  }

  public void setShipMass( double mass ) {
    shipMass = mass;
  }

  public void setMaxSpeed( double speed ) {
    MAX_VEL = speed;
  }

	public int getX() {
		return x;
	}

	public int getY() {
		return y;
	}

	public void restart() {
    sprite = new Polygon();
		this.x = standardX;
		this.y = standardY;
		vel_x = vel_y = 0D;
		deltaVel = 0D;
		if( x < maxX/2 ) angolo = 0D;
		else angolo = Math.PI;
	}

	public void ruota(){
		angolo += rotationDelta;
		if( angolo >= 2*Math.PI ) angolo -= 2*Math.PI;
		else if( angolo <= 0 ) angolo += 2*Math.PI;
	}

  public void setRotationDelta( double ang ) {
    rotationDelta = ang;
  }

	public void accelera( double force ) {
		if( force < MAX_ACC && force > -MAX_ACC )
     		deltaVel = force/shipMass;
	}

	public double getAngle() {
		return angolo;
	}

  public double getSpeedX() {
    return vel_x;
  }

  public double getSpeedY() {
    return vel_y;
  }

	public Astronave[] getExplosion() {
    Astronave[] a = new Astronave[shape.npoints];
    for( int i = 0 ; i < shape.npoints ; i++ ) {
      Polygon p = new Polygon();
      p.addPoint( shape.xpoints[i] , shape.ypoints[i] );
      p.addPoint( shape.xpoints[((i+1)>(shape.npoints-1))?0:(i+1)] , shape.ypoints[((i+1)>(shape.npoints-1))?0:(i+1)] );
      p.addPoint( 0 , 0 );
      a[i] = new Astronave( p , x , y , fill);
      a[i].setGravityCenter( centroX , centroY );
      a[i].setArea( maxX , maxY );
      a[i].setShipMass( shipMass/shape.npoints );
      a[i].setRotationDelta( Math.random()*Math.PI/5D );
      a[i].angolo = Math.random()*2D*Math.PI;
      a[i].vel_x = vel_x + 5*Math.cos(a[i].angolo);
      a[i].vel_y = vel_y + 5*Math.sin(a[i].angolo);
    }
    return a;
	}

	public void sposta() {
		double distY = centroY-y, distX = x-centroX;
		double dist = Math.sqrt(Math.pow((distY),2) + Math.pow((distX),2));


    // Rotazione
    ruota();

		// Accelerazione
		// Effetto motori
		vel_x += deltaVel*Math.cos(angolo);
		vel_y += deltaVel*Math.sin(angolo);

		// Effetto attrito
		// vel_x -= attr*vel_x;
		// vel_y -= attr*vel_y

		// Effetto gravita'
		if( dist != 0 ) {
			vel_x -= shipMass*grav*( (distX) / dist )*Math.pow((dist),-2);
			vel_y -= shipMass*grav*( (distY) / dist )*Math.pow((dist),-2);
		}


		if( Math.abs(vel_x) > MAX_VEL ) vel_x = (vel_x>0)?MAX_VEL:-MAX_VEL;
		if( Math.abs(vel_y) > MAX_VEL ) vel_y = (vel_y>0)?MAX_VEL:-MAX_VEL;

		// Spostamento
		x += Math.round(vel_x);
		y -= Math.round(vel_y);


		// Traslazione dopo i bordi
		reset();
	}


	public void reset() {
		// Traslazione dopo i bordi
		if( x > maxX ) { x = 0 ;}
		if( y > maxY ) { y = 0 ;}
		if( x < 0 ) { x = maxX;}
		if( y < 0 ) { y = maxY;}
	}

	public void draw( Graphics g ) {
		sprite = new Polygon();
		for (int i = 0; i < this.shape.npoints; i++)
			sprite.addPoint( (int) Math.round( shape.xpoints[i]*Math.cos(angolo) + shape.ypoints[i]*Math.sin(angolo) +x),
							 (int) Math.round( shape.ypoints[i]*Math.cos(angolo) - shape.xpoints[i]*Math.sin(angolo) +y) );

		if( !fill) g.drawPolygon( sprite );
		else g.fillPolygon( sprite );
	}

	public boolean isColliding( Astronave a ) {
		for( int i=0 ; i< this.sprite.npoints ; i++)
			if( a.sprite.contains( this.sprite.xpoints[i] , this.sprite.ypoints[i] ) ) return true;
    for( int i=0 ; i< this.sprite.npoints ; i++)
			if( this.sprite.contains( a.sprite.xpoints[i] , a.sprite.ypoints[i] ) ) return true;
		return false;
	}

  public void bounce( Astronave a ) {
    double i;
    i = this.vel_x;
    this.vel_x = a.vel_x;
    a.vel_x = i;
    i = this.vel_y;
    this.vel_y = a.vel_y;
    a.vel_y = i;
  }
}

class Shot extends Astronave
{
	boolean active = false;

	int lifespan = 100;

	double VEL_SHOT = 15D;

	public Shot( int x , int y , double angle , double speedX , double speedY ) {
		super( x , y , false );
    MAX_VEL = 100D;
    shipMass = 0.5;
		angolo = angle;
		vel_x = speedX + VEL_SHOT*Math.cos(angolo);
		vel_y = speedY + VEL_SHOT*Math.sin(angolo);
		shape = new Polygon();
		shape.addPoint(0,0);
		sprite = new Polygon();
		sprite.addPoint(x,y);
		active = true;

	}

	public void reset() {
		if( x > maxX || y > maxY || x < 0 || y < 0 || --lifespan == 0 )
			active = false;
	}

	public boolean getStatus() {
		return active;
	}

	public void setStatus( boolean active ) {
		this.active = active;
	}

	public void draw( Graphics g ) {
		sprite = new Polygon();
		sprite.addPoint( x-4 , y-4 );
    sprite.addPoint( x-4 , y+4 );
    sprite.addPoint( x+4 , y-4 );
    sprite.addPoint( x+4 , y+4 );
		g.fillOval( x-4 , y-4 , 8 , 8 );
	}
};

public class AstroApplet extends Applet implements Runnable,KeyListener
{
	// Massimo numero di colpi sparabili
	final int MAX_SHOT = 10;

	// Colpo sparato correntemente
	int currShot = 0 , currShot1=0;

	// Init dell'oggetto astronave
	Astronave ship = new Astronave(60,240,true);
	Astronave ship1 = new Astronave(580,240,false);

	// Array di colpi (massimo numero di colpi sparati)
	Shot[] shot1 = new Shot[MAX_SHOT];
  Shot[] shot2 = new Shot[MAX_SHOT];

  // Init delle esplosioni
  Astronave[] explosion,explosion1;
  int explosionCount1 = 0 , explosionCount = 0;
  static final int PLAYER1 = 0;
  static final int PLAYER2 = 1;
  static final int PLAYER_BOTH = 2;

  // Parametri delle astronavi (potenza dei motori, massa, ecc)
  double THRUST_POWER = 1D;
  double BACK_THRUST_POWER = 0.5D;
  double SHIP_MASS = 1D;
  double MAX_SPEED = 10D;

  // Parametri dell'universo (gravita', dimensioni della stella, colore, ecc.)
  double STAR_MASS = 2500D;
  int STAR_DIM = 15;
  Color STAR_COLOR = Color.yellow;

  // Finestra per l'immissione dei parametri
  ParameterFrame parameters = new ParameterFrame();

	// Ciclo principale
	Thread ciclo;
  Intro intro;
  Applet questo = this;
  boolean paused = false , started = false;

 	// OffScreen buffer
	Image offScreenImage;
	Graphics offScreen;

	public void init() {
   parameters.setThrustPower( THRUST_POWER );
   parameters.setBackThrustPower( BACK_THRUST_POWER);
   parameters.setShipMass( SHIP_MASS );
   parameters.setGravity( STAR_MASS );
   parameters.setSpeed( MAX_SPEED );

   parameters.ok.addActionListener( new ActionListener() {
      public void actionPerformed( ActionEvent e ) {
        THRUST_POWER = parameters.getThrustPower();
        BACK_THRUST_POWER = parameters.getBackThrustPower();
        ship.setMaxSpeed(MAX_SPEED = parameters.getSpeed());
        ship1.setMaxSpeed(MAX_SPEED);
        ship.setShipMass( SHIP_MASS = parameters.getShipMass() );
        ship1.setShipMass( SHIP_MASS );
        Astronave.setCenterMass( parameters.getGravity() );
        parameters.setVisible(false);
        unpause();
      }});

    parameters.addWindowListener( new WindowAdapter() {
      public void windowClosing( WindowEvent e ){
        parameters.setVisible(false);
        unpause();
      }});

    // Init dei Listener della tastiera
		addKeyListener(this);

		requestFocus();
		setBackground( Color.black );
		setForeground( Color.white );

		offScreenImage = createImage( getSize().width , getSize().height );
		offScreen = offScreenImage.getGraphics();

		for( int i = 0 ; i < MAX_SHOT ; i++ )
			shot1[i] = null;
	}

 	public void start() {
		if( intro == null ) {
      intro = new Intro( this , offScreen );
  		intro.start();
		}
    if ( ciclo == null ) {
      ciclo = new Thread(this);
    }
	}

	public void stop() {
		ciclo = null;
    intro = null;
    started = false;
	}

	public void run() {
		Thread me = Thread.currentThread();
		while( me == ciclo ) {
			// Aggiorno il movimento con le equazioni del moto
			if( explosionCount <= 0 ) ship.sposta();
      else {
        for( int i = 0 ; i < explosion.length ; i++ )
          explosion[i].sposta();
        explosionCount--;
      }

			if( explosionCount1 <= 0 ) ship1.sposta();
      else {
        for( int i = 0 ; i < explosion1.length ; i++ )
          explosion1[i].sposta();
        explosionCount1--;
      }

			// Aggiorno il movimento e il fuoco dei colpi
			updateShots();

			// Verifico che nessuna nave sia stata colpita
			checkStatus();

			// Ridisegno lo schermo
			redraw();
			try
			{
				Thread.sleep(30);
        synchronized( this ) {
           while( paused ) wait();
        }
			}
			catch ( InterruptedException e )
			{
			}
		}
	}

  public void explode( int i ){
    switch(i) {
    case AstroApplet.PLAYER1 :
        explosion = ship.getExplosion();
        explosionCount = 30;
        ship.restart(); break;
    case AstroApplet.PLAYER2 :
        explosion1 = ship1.getExplosion();
        explosionCount1 = 30;
        ship1.restart(); break;
    default:
        explosion = ship.getExplosion();
        explosionCount = 30;
        explosion1 = ship1.getExplosion();
        explosionCount1 = 30;
        explosionCount1 = explosionCount = 30;
        ship1.restart();
        ship.restart();
    }
  }

	public void checkStatus() {
    double centroX,centroY;
    centroX = getSize().width/2;
    centroY = getSize().height/2;

		// Astronavi
		// Verifica che non finiscano dentro il sole
    if( Math.abs(ship.getX()-centroX) > STAR_DIM ||
        Math.abs(ship.getY()-centroY) > STAR_DIM ) {
      for( int i = 0 ; i < ship.sprite.npoints ; i++ )
	    	if( Math.abs(ship.sprite.xpoints[i]-getSize().width/2)  < STAR_DIM &&
		        Math.abs(ship.sprite.ypoints[i]-getSize().height/2) < STAR_DIM ) explode(PLAYER1);
    } else explode(PLAYER1);

    if( Math.abs(ship1.getX()-centroX) > STAR_DIM ||
        Math.abs(ship1.getY()-centroY) > STAR_DIM ) {
        for( int i = 0 ; i < ship1.sprite.npoints ; i++ )
      		if( Math.abs(ship1.sprite.xpoints[i]-getSize().width/2)  < STAR_DIM &&
		          Math.abs(ship1.sprite.ypoints[i]-getSize().height/2) < STAR_DIM ) explode(PLAYER2);
    } else explode(PLAYER2); 

    // Verifica che le astronavi non si tocchino
    if( ship1.isColliding( ship ) ) {
      explode(PLAYER_BOTH);
    }

		// Proiettili
    // Verifica che non siano state colpite
		for( int i = 0 ; i < MAX_SHOT ; i++ ) {
		if( shot1[i] != null && shot1[i].getStatus() && shot1[i].isColliding(ship1) ) {
			explode(PLAYER2);
			shot1[i].setStatus(false);
		} else if( shot1[i] != null && shot1[i].getStatus() && Math.abs(shot1[i].getX()-getSize().width/2)  < STAR_DIM &&
		    Math.abs(shot1[i].getY()-getSize().height/2) < STAR_DIM ) shot1[i].setStatus(false);
		if( shot2[i] != null && shot2[i].getStatus() && shot2[i].isColliding(ship) ) {
			explode(PLAYER1);
			shot2[i].setStatus(false);
		} else if( shot2[i] != null && shot2[i].getStatus() && Math.abs(shot2[i].getX()-getSize().width/2)  < STAR_DIM &&
		    Math.abs(shot2[i].getY()-getSize().height/2) < STAR_DIM ) shot2[i].setStatus(false);
		}
	}

	public void updateShots() {

		for( int i = 0 ; i < MAX_SHOT ; i++ ) {
			if( shot1[i] != null && shot1[i].getStatus() ) {
				shot1[i].sposta();
			}
      if( shot2[i] != null && shot2[i].getStatus() ) {
				shot2[i].sposta();
			}
    }
	}

	public void drawShots(Graphics g) {
		for( int i = 0 ; i < MAX_SHOT ; i++ ) {
			if( shot1[i] != null && shot1[i].getStatus() ) {
        g.setColor( Color.red );
				shot1[i].draw(g);
      }
      if( shot2[i] != null && shot2[i].getStatus() ) {
        g.setColor( Color.green );
				shot2[i].draw(g);
      }
    }
    g.setColor( Color.white );
	}

	public void update( Graphics g ) {
		paint(g);
	}

	public void paint( Graphics g ) {

		// Copia sulla finestra principale
	  g.drawImage( offScreenImage , 0 , 0, this );
	}

	public void redraw() {
		offScreen.clearRect( 0 , 0, getSize().width , getSize().height );

		// Astronavi
		offScreen.setColor( Color.white );
		if( explosionCount <= 0 ) ship.draw(offScreen);
    else
      for( int i = 0 ; i < explosion.length ; i++ ){
/*        if(Math.abs(explosion[i].getX()-getSize().width/2) > 15 &&
           Math.abs(explosion[i].getY()-getSize().height/2) > 15 ) */
                explosion[i].draw(offScreen);
      }


		if( explosionCount1 <= 0 ) ship1.draw(offScreen);
    else
      for( int i = 0 ; i < explosion1.length ; i++ ) {
/*        if(Math.abs(explosion1[i].getX()-getSize().width/2) > 15 &&
           Math.abs(explosion1[i].getY()-getSize().height/2) > 15 )  */
                explosion1[i].draw(offScreen);
      }

		// Proiettili
		drawShots(offScreen);

		// Stella
		offScreen.setColor( STAR_COLOR );
		offScreen.fillOval( getSize().width/2 -STAR_DIM , getSize().height/2 -STAR_DIM , STAR_DIM*2 , STAR_DIM*2 );
		repaint();
	}

	public void keyPressed( KeyEvent e ) {
 	int i = e.getKeyCode();
		switch( i ) {

			// Nave 1
			case KeyEvent.VK_UP : ship.accelera( THRUST_POWER ); break;
			case KeyEvent.VK_DOWN : ship.accelera( -BACK_THRUST_POWER ); break;
			case KeyEvent.VK_LEFT : ship.setRotationDelta( Math.PI*(1D/10D) ); break;
			case KeyEvent.VK_RIGHT : ship.setRotationDelta( -Math.PI*(1D/10D) ); break;

		  // Nave 2
      case KeyEvent.VK_W : ship1.accelera( THRUST_POWER ); break;
			case KeyEvent.VK_S : ship1.accelera( -BACK_THRUST_POWER ); break;
			case KeyEvent.VK_A : ship1.setRotationDelta( Math.PI*(1D/10D) ); break;
			case KeyEvent.VK_D : ship1.setRotationDelta( -Math.PI*(1D/10D) ); break;
			default :
    }
	}

	public void keyReleased( KeyEvent e ) {
		int i = e.getKeyCode();
		switch( i ) {
      case KeyEvent.VK_ENTER : intro.end() ; ciclo.start(); break;

			// Nave 1
			case KeyEvent.VK_UP : ship.accelera( 0D ); break;
			case KeyEvent.VK_DOWN : ship.accelera( 0D ); break;
      case KeyEvent.VK_LEFT : ship.setRotationDelta( 0D ); break;
			case KeyEvent.VK_RIGHT : ship.setRotationDelta( 0D ); break;

      case KeyEvent.VK_0:
			case KeyEvent.VK_NUMPAD0 :
        if( explosionCount > 0 ) break;
				shot1[currShot++] = new Shot( ship.getX() , ship.getY() , ship.getAngle() ,
                                      ship.getSpeedX() , ship.getSpeedY());
				if( currShot >= MAX_SHOT ) currShot = 0; break;

			// Nave 2
  		case KeyEvent.VK_W : ship1.accelera( 0D ); break;
	  	case KeyEvent.VK_S : ship1.accelera( 0D ); break;
      case KeyEvent.VK_A : ship1.setRotationDelta( 0D ); break;
		  case KeyEvent.VK_D : ship1.setRotationDelta( 0D ); break;

      case KeyEvent.VK_SPACE :
        if( explosionCount1 > 0 ) break;
				shot2[currShot1++] = new Shot( ship1.getX() , ship1.getY() , ship1.getAngle() ,
                                       ship1.getSpeedX() , ship1.getSpeedY());
				if( currShot1 >= MAX_SHOT ) currShot1 = 0; break;


      case KeyEvent.VK_P : paused = true ; parameters.setVisible(true); break;

		  default:
		}
	}

	public void keyTyped( KeyEvent e ) {
	}

  private void pause() {
    paused = true;
  }

  private synchronized void unpause() {
    paused = false;
    notify();
  }
}

class Intro extends Thread {

  Container a;
  Graphics offScreen;
  boolean stopped = false;

  String[] frasi =
    { "...Gravity Wars..." ,
      "...a game brought you by..." ,
      "...Giovanni Beltrame..." ,
      "...click or press a key to start..." };

  public Intro( Container cont , Graphics buffer ) {
    a = cont;
    offScreen = buffer;
  }

  public void end() {
    stopped = true;
  }

  public void begin() {
    stopped = false;
  }

  public void run() {
    Thread me = Thread.currentThread();

    Fader f = new Fader();
    for( int i = 0 ; i < frasi.length ; i++ )
      f.addText(  frasi[i] , new Font( "Tahoma" , Font.PLAIN , 32 ) ,
                  null , a.createImage(1,1) );
    f.setDelta(4);

    int pausa;
    while( !stopped ) {
      pausa = f.changeColors();
      f.draw( a , offScreen );
      try {
        Thread.sleep(pausa*10);
      } catch( InterruptedException e ) {}
        a.repaint();
    }
  }

}

